#!/usr/bin/env python3

# Copyright (c) 2023, harens
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of seaport nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Python API for MacPorts portfiles."""

import re
import subprocess
import sys
from typing import Optional

from beartype import beartype
from beartype.typing import List, Tuple

from seaport._clipboard.format import format_subprocess

# Don't count code coverage since different python versions
# won't run different parts of code
if sys.version_info >= (3, 8):  # pragma: no cover
    from beartype.typing import Final
else:  # pragma: no cover
    from typing_extensions import Final


# TODO: Set no output (especially for errors)
class Port:
    """Scrapes portfile info for usage in Python modules.

    Examples:
        >>> from seaport.portfile import Port
        >>> port = Port("py-base91")
        >>> port.version
        '1.0.1'

        >>> from seaport.portfile import Port
        >>> port = Port("py-base91")
        >>> port.revision
        0

        >>> from seaport.portfile import Port
        >>> try:
        ...     port = Port("non-existent-port")
        ... except Exception:
        ...     pass
        >>> # This raises an exception

    Attributes:
        name (str): The name of the port e.g. gping
    """

    @beartype
    def __init__(self, name: str) -> None:
        """Set optional attributes and check if port exists."""
        # TODO: Figure out how to find path without subprocess
        # TODO: Refactor this
        # TODO: This also kind of defeats the purpose of that bandit error, so find a better way to determine the path
        # no forward slash at end for Bandit B607
        self.name: Final[str] = name
        self._path: Final[str] = format_subprocess(["/usr/bin/which", "port"]).replace(
            "/port", ""
        )

        try:
            self._info: Final[str] = format_subprocess(
                [f"{self._path}/port", "info", self.name]
            )
        except subprocess.CalledProcessError:
            # TODO: Set a more specific exception
            raise Exception(f"{self.name} doesn't exist, run portindex if port is new")

        self._parsedInfo: Final[List[str]] = self._info[
            self._info.find("@") + 1 :
        ].split()
        versionParse: Final[List[str]] = self._parsedInfo[0].split("_")

        # Parse saved port info, falling back to calling the explicit function
        # As a quick sanity check, see that the first digit of the version number is indeed a digit

        self.version: Final[str] = (
            format_subprocess(
                [f"{self._path}/port", "info", "--version", self.name]
            ).split(" ")[1]
            if len(versionParse) not in (1, 2) or not versionParse[0][0].isdigit()
            else versionParse[0]
        )

        self.revision: Final[int] = (
            int(
                format_subprocess(
                    [f"{self._path}/port", "info", "--revision", self.name]
                ).split(" ")[1]
            )
            if len(versionParse) not in (1, 2) or not versionParse[0][0].isdigit()
            else 0
            if len(versionParse) == 1
            else int(versionParse[1])
        )

    @beartype
    def __str__(self) -> str:
        """Outputs the name and version of the port.

        Examples:
            >>> from seaport.portfile import Port
            >>> print(Port("py-base91"))
            py-base91 1.0.1
        """
        return f"{self.name} {self.version}"

    @beartype
    def __repr__(self) -> str:
        """Outputs the attributes that a port was defined with.

        Examples:
            >>> from seaport.portfile import Port
            >>> Port("py-base91")
            Port(name=py-base91)
        """
        return f"Port(name={self.name})"

    # TODO: These livecheck tests will fail when I least expect it
    @beartype
    def livecheck(self) -> str:
        """Runs port livecheck to check for any new versions.

        If no livecheck is available or the portfile is already the latest version, the current version is outputted.
        Note that this can be slow, since it's scraping `port livecheck`.

        Examples:
            >>> from seaport.portfile import Port
            >>> port = Port("py-base91")
            >>> port.livecheck()
            '1.0.1'

            >>> from seaport.portfile import Port
            >>> port = Port("py39-base91")
            >>> port.livecheck()
            '1.0.1'

        Returns:
            A string representing the latest version.
        """
        # Take the last word of port livecheck, and then remove the bracket
        update = format_subprocess(
            [f"{self._path}/port", "livecheck", self.name]
        ).split(" ")[-1][:-1]

        # If there's no livecheck output, fallback to subport
        # Convoluted if statement to make mypy happy
        if update == "":
            # Makes mypy happy since a function could theoretically change to be None
            subports = self.subports()
            if subports is not None:
                update = format_subprocess(
                    [f"{self._path}/port", "livecheck", subports[-1]]
                ).split(" ")[-1][:-1]

        # If there's no livecheck output again, fallback to current version
        # Implies no livecheck available or already up-to-date
        # N.B. str is required for py 3.7 type checking
        return update if update != "" else str(self.version)

    @beartype
    def subports(self) -> Optional[List[str]]:
        """Determines a list of subports of a port.

        If there are no subports available, None is outputted.

        Examples:
            >>> # Subports available
            >>> from seaport.portfile import Port
            >>> port = Port("py-base91")
            >>> port.subports()
            ['py38-base91', 'py39-base91']

            >>> # Subports not available
            >>> from seaport.portfile import Port
            >>> port = Port("folderify")
            >>> print(port.subports())
            None

        Returns:
            A list representing all the subports of the port.
        """
        # Split subport section by colon and comma
        # This needs to be made more efficient
        # TODO: Refactor so that as soon as a line containing sub-ports is found, don't bother
        # with other iterations in for list comprehension
        return (
            None
            if "Sub-ports" not in self._info
            else [
                i.replace(" ", "")
                for i in re.split(
                    "[:,]",
                    " ".join([s for s in self._info.splitlines() if "Sub-ports" in s]),
                )
                if i != "Sub-ports"
            ]
        )

    @beartype
    def checksums(self, _name: Optional[str] = None) -> Tuple[str, str, str, str]:
        """Determines the current checksums of a portfile.

        For python ports, their pyXY- subport is used to determine the checksums.
        Note that this method only works for ports with the standard rmd/sha/size setup (not the older format), and it can also be
        quite slow since it's scraping `port distfiles NAME`

        Examples:
            >>> # Determines rmd160/sha256/size/website
            >>> from seaport.portfile import Port
            >>> port = Port("py-base91")
            >>> port.checksums()
            ('c1bd97759a8d7bfdb95cd76ada05efa9e9d99f28', '5b284a2ba3c97be1eb9473f3af94a9bf141d61005d836e75e645d2798da58799', '2331', 'https://files.pythonhosted.org/packages/source/b/base91/base91-1.0.1.tar.gz')


        Returns:
            rmd160, sha256, size and the website that provided the distfile.
        """
        # Name is used if recursion required for subports
        _name = self.name if _name is None else _name
        distfiles = (
            format_subprocess([f"{self._path}/port", "distfiles", _name])
            .replace("\n ", "")
            .split(" ")
        )
        try:
            # We're only interested in the first result
            # Credit to https://stackoverflow.com/a/9868665/10763533
            website: Final[str] = next(
                s for s in distfiles if "http://" in s or "https://" in s
            )
        except StopIteration:
            # Tries to determine the subport
            # This is since the distfiles cmd only works for subports
            subports = self.subports()
            if subports is None:
                raise Exception(f"port distfiles {_name} provides no output")
            # Repeat the process with the subport
            return self.checksums(subports[-1])

        website_index: Final[int] = distfiles.index(website)

        # rmd, sha, size, download website
        # TODO: This will not work for the old format
        return (
            distfiles[website_index - 3][:-7],
            distfiles[website_index - 2][:-5],
            distfiles[website_index - 1],
            website,
        )

    @beartype
    def primary_category(self) -> str:
        """Determines the first category of a port.

        This is useful to determine which folder of the macports repo the port would reside in.

        Examples:
            >>> from seaport.portfile import Port
            >>> port = Port("gping")
            >>> port.primary_category()
            'net'

        Returns:
            The category of the port e.g. sysutils.
        """
        # Remove leftmost bracket and rightmost comma (if multiple categories) or bracket (if only one)
        if self._parsedInfo[1][0] != "(" or self._parsedInfo[1][-1] not in (
            ")",
            ",",
        ):
            category_list: Final[List[str]] = format_subprocess(
                [f"{self._path}/port", "info", "--category", self.name]
            ).split(" ")
            # Remove comma, and only take the first category
            # N.B. str seems to be required for py37 type checking
            return (
                str(category_list[1][:-1])
                if len(category_list) > 2
                else str(category_list[1])
            )
        else:
            # N.B. str is required for python type checking
            return str(self._parsedInfo[1][1:-1])



#!/usr/bin/env python3

# Copyright (c) 2023, harens
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of seaport nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Additional checks that aren't provided by default.

e.g. Linting
"""

import subprocess
from typing import Optional

import click
from beartype import beartype

from seaport._clipboard.checks import user_path
from seaport._clipboard.format import format_subprocess


@beartype
def perform_lint(name: str) -> bool:
    """Lints the port and checks output for errors.

    Args:
        name: The name of the port

    Returns:
        bool: Whether the linting was successful or not
    """
    click.secho("🤔 Linting", fg="cyan")
    lint_output = format_subprocess(
        [f"{user_path(True)}/port", "lint", "--nitpick", name]
    )
    click.echo(lint_output)
    output_list = lint_output.split(" ")

    # Finds the no. of errors and warnings
    errors = int(output_list[output_list.index("errors") - 1])
    warnings = int(output_list[output_list.index("warnings") - 1])

    if errors > 1:
        # Fail if there are any errors
        return False
    if warnings > 1:
        # Ask whether the user wishes to continue
        if not click.confirm(
            f"There are {warnings} warnings. Do you wish to continue?"
        ):
            return False
    return True


@beartype
def perform_test(name: str, subport: Optional[str] = None) -> bool:
    """Tests the port and checks output for errors.

    Args:
        name: The name of the port
        subport: The name of one of the subports

    Returns:
        bool: Whether the tet was successful or not
    """
    click.secho(f"🧪 Testing {name}", fg="cyan")
    try:
        subprocess.run(
            [f"{user_path()}/sudo", f"{user_path(True)}/port", "test", name],
            check=True,
        )
    except subprocess.CalledProcessError:
        # For python ports, the tests are in the subport
        # There are no tests in the original port
        if subport:
            click.secho(f"🏗 Trying with subport {subport}", fg="cyan")
            try:
                subprocess.run(
                    [f"{user_path()}/sudo", f"{user_path(True)}/port", "test", subport],
                    check=True,
                )
            except subprocess.CalledProcessError:
                click.secho("❌ Tests failed", fg="red")
                return False
        else:
            click.secho("❌ Tests failed", fg="red")
            return False
    click.secho("✅ Tests passed", fg="green")
    return True


@beartype
def perform_install(name: str) -> None:
    """Runs sudo port -vst install NAME.

    Args:
        name: The name of the port
    """
    click.secho(f"🏗️ Installing {name}", fg="cyan")
    subprocess.run(
        [
            f"{user_path()}/sudo",
            f"{user_path(True)}/port",
            "-vt",
            "install",
            name,
        ],
        check=True,
    )
    click.secho(
        "Paused to allow user to test basic functionality in a different terminal",
        fg="cyan",
    )
    if click.confirm("Do you want to uninstall the port?"):
        click.secho(f"🗑  Uninstalling {name}", fg="cyan")
        subprocess.run(
            [f"{user_path()}/sudo", f"{user_path(True)}/port", "uninstall", name],
            check=True,
        )



#!/usr/bin/env python3

# Copyright (c) 2023, harens
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of seaport nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""The main CLI function, which the user runs."""

import os
import subprocess
import sys
import tempfile
from typing import Optional

import click
from beartype import beartype

from seaport._click_functions import main_cmd
from seaport._clipboard.additional import perform_install, perform_lint, perform_test
from seaport._clipboard.checks import user_path
from seaport._clipboard.portfile.checksums import new_checksums, replace_checksums
from seaport._clipboard.portfile.portfile_numbers import new_version
from seaport._clipboard.user import clean, user_clipboard
from seaport.portfile import Port


# Some parameters are not used
# They are only here since the pr function sends them
@click.command()
@main_cmd
@beartype
def clip(
    name: str,
    bump: Optional[str],
    test: bool,
    lint: bool,
    url: Optional[str],
    install: bool,
    write: bool,
    location: Optional[str] = None,
    new: bool = False,
) -> None:
    """Bumps the version number and checksum of NAME.

    It then copies the result to your clipboard.
    """
    # Tasks that definitely require sudo
    sudo = test or install

    port = Port(name)

    old_checks = port.checksums()

    # Determine new version
    bump = new_version(port, bump)

    click.secho(f"👍 New version is {bump}", fg="green")

    # Allows pr function to get the version number and category
    # new_version checks if bump is none and deals with it there
    os.environ["BUMP"] = bump
    os.environ["CATEGORY"] = port.primary_category()

    # Allows setting custom url
    new_website = old_checks[3].replace(port.version, bump) if url is None else url

    # Parameter is new website (old website with old version replaced with new version)
    new_sha256, new_rmd160, new_size = new_checksums(new_website)

    click.secho("🔎 Checksums:", fg="cyan")
    click.echo(f"Old rmd160: {old_checks[0]}")
    click.echo(f"New rmd160: {new_rmd160}")
    click.echo(f"Old sha256: {old_checks[1]}")
    click.echo(f"New sha256: {new_sha256}")
    click.echo(f"Old size: {old_checks[2]}")
    click.echo(f"New size: {new_size}")

    # Add the new checksums, and take a backup of the original
    file_location = (
        subprocess.check_output([f"{user_path(True)}/port", "file", name])
        .decode("utf-8")
        .strip()
    )

    with click.open_file(file_location) as file:
        # Backup of the original contents
        original = file.read()

    new_contents = replace_checksums(
        original,
        (old_checks[0], old_checks[1], old_checks[2], port.version),
        (new_rmd160, new_sha256, new_size, bump),
    )

    if sudo or write or lint:
        # Temporary files created to get around sudo write problem
        tmp_version = tempfile.NamedTemporaryFile(mode="w")
        tmp_version.write(new_contents)
        tmp_version.seek(0)

        click.secho("💾 Editing local portfile repo", fg="cyan")
        if not write:
            # Changes only reverted if the user doesn't use the --write flag
            click.secho("📝 Changes will be reverted after completion", fg="cyan")

        subprocess.run(
            ([] if os.access(file_location, os.W_OK) else [f"{user_path()}/sudo"])
            + ["cp", tmp_version.name, file_location],
            check=True,
        )

        if write:
            click.secho(
                "📝 The portfile's contents have been updated",
                fg="cyan",
            )

        if lint:
            # If the lint is not successful
            if not perform_lint(name):
                clean(original, file_location, name)
                sys.exit(1)

        if test:
            subport = port.subports()
            if subport is not None:
                result = perform_test(name, subport[-1])
            else:
                result = perform_test(name)
            # If the tests fail
            if not result:
                clean(original, file_location, name)
                sys.exit(1)

        if install:
            perform_install(name)

        clean(original, file_location, name, write, sudo)

    # Clipboard functions at the very end
    # to reduce the chance of user's clipboard being changed
    # after adding contents
    user_clipboard(new_contents)
